#! /usr/bin/env python

import sys
from LN import LN, MLLN
import Node

#>=== Match ======================================================

RAISE = True
NORAISE = False

# If I ever cache Matches, the following will have to be done:
# * Associate a pattern with each Match.
# * Define __eq__ for patterns.
# * Design null patterns to wrap around patterns in matcher, etc. The pattern
#   that is stored must contain the full history of the Match.
# * Merge BaseChainPattern and ChildPattern.

class FailedMatchError(Exception):
    """Raised if a match fails"""
    pass

class Match(object):
    def __init__(self, startnode, endnodes):
        if isinstance(endnodes, Node.Node):
            endnodes = (endnodes,)
        if not isinstance(endnodes, (list, tuple)):
            raise TypeError, 'second argument must be a Node, list or tuple'
        self.startnode = startnode
        self.endnodes = endnodes
        self.supertree = self.startnode.tree
        for node in self.endnodes:
            if node.tree != self.supertree:
                raise TypeError, 'endnodes is must be in the same tree as teh startnode'

    # Create ".classname" attribute for this class and all its descendents.
    def get_classname(self):
        return self.__class__.__name__
    classname = property(get_classname)

    def __str__(self):
        parts = ['<%s:' % self.classname]
        parts.append('  startnode:')
        startpart = '    ' + str(self.startnode).replace('\n', '\n    ')
        parts.append(startpart)
        parts.append('  endnodes:')
        for endnode in self.endnodes:
            endnodepart = '    ' + str(endnode).replace('\n', '\n    ')
            parts.append(endnodepart)
        parts[-1] = parts[-1] + '>'
        return '\n'.join(parts)

    def __repr__(self):
        parts = ['<%s:' % self.classname]
        parts.append('  startnode:')
        startpart = '    ' + repr(self.startnode).replace('\n', '\n    ')
        parts.append(startpart)
        parts.append('endnodes:')
        for endnode in self.endnodes:
            endnodepart = '    ' + repr(endnode).replace('\n', '\n    ')
            parts.append(endnodepart)
        parts[-1] = parts[-1] + '>'
        return '\n'.join(parts)

    def merge_below(self, matches):
        """Merge "matches" onto the endnodes of "self".

        The startnode for each match in "matches' must be an endnode of
        "self".

        If your program is SLOW, here is a good place to look. Note the
        self.supertree.walk() below. If you plan to merge many matches, put
        them in a list and call one of the merge fucntions once.
        """
        if isinstance(matches, Match):
            matches = (matches,)
        for match in matches:
            if not isinstance(match, Match):
                raise TypeError, 'argument is not a Match'
            if match.startnode not in self.endnodes:
                raise TypeError, 'each startnode of the matches must be ' \
                    'an endnode of self'
        return merge(self.startnode, matches)

    def get_success(self):
        return not isinstance(self, FailedMatch)
    # "success" is used like a attribute of Match.
    success = property(get_success)

    def get_failure(self):
        return isinstance(self, FailedMatch)
    # "failure" is used like a attribute of Match.
    failure = property(get_failure)

    def the_endnode(self):
        if len(self.endnodes) == 1:
            return self.endnodes[0]
        else:
            raise ValueError, 'length of endnodes must be one'

    def get_tokens(self):
        out = []
        for endnode in self.endnodes:
            tokens = endnode.get_tokens()
            out.extend(tokens)
        return out

def merge(topstartnode, matches):
    """Merge a list of matches under topstartnode.

    If your program is SLOW, here is a good place to look. Note the "walk()"
    below. If you plan to merge many matches, put them in a list and call
    one of the merge functions once.
    """
    # Here are requirements for "merge":
    # 1. "topstartnode.tree" must the supertree for all the matches.
    # 2. All the startnodes in the matches must be below topstartnode in the
    #    supertree.
    # 3. The startnodes must all be different from each other.
    # 4. The startnode of the merged match is "topstartnode".
    # 5. Each endnode of the merged match must be an endnode of the input
    #    matches. But no endnode of the merged match can have another
    #    endnode below it.

    if isinstance(matches, Match):
        matches = (matches,)
    if len(matches) == 0:
        return EmptyMatch()
    if len(matches) == 1:
        match = matches[0]
        if not isinstance(match, Match):
            raise TypeError, 'argument is not a Match'
        if topstartnode == match.startnode:
            return match

    # Collect the startnodes and endnodes without redundency. Check the
    # requirements.
    startnodedict = dict()
    startnodedict[topstartnode.id] = topstartnode
    endnodedict = dict()
    for match in matches:
        if not isinstance(match, Match):
            raise TypeError, 'argument is not a Match'
        startnodedict[match.startnode.id] = match.startnode
        if match.supertree != topstartnode.tree:
            raise TypeError, 'the supertree for each match must be "topstartnode.tree"'
        for endnode in match.endnodes:
            endnodedict[endnode.id] = endnode

    # Reverse walk the part of the supertree under "topstartnode". This is a
    # bottom-up walk. During this walk, each node is given a new attribute
    # "node_ok". A node is ok if none of the nodes _strictly_ below it are
    # endnodes of the input matches.

    newendnodesdict = dict()
    newendsnodelist = list()
    startnodes_found = dict()
    for node in topstartnode.reverse_walk():
        if node.id in startnodedict:
            startnodes_found[node.id] = None
        if node.isleaf():
            # A leaf cannot have any endnodes strictly below it.
            node.node_ok = True
        else:
            # A node is ok if:
            #    Its children have no endnodes strictly below them.
            # and
            #    None of its children are endnodes.
            node.node_ok = True
            for child in node.children:
                node.node_ok = (node.node_ok and child.node_ok and
                        not child.id in endnodedict)

        if node.id in endnodedict and node.node_ok:
            newendnodesdict[node.id] = None
            newendsnodelist.append(node)
        # Do not want to keen the "node_ok" attribute.
        for child in node.children:
            del child.node_ok
    del topstartnode.node_ok
    newendsnodelist.reverse()

    for id in startnodedict:
        if id not in startnodes_found:
            raise ValueError, 'startnode for a match not under topstartnode'

    # Debugging check.
    for node in topstartnode.walk():
        if hasattr(node, 'node_ok'):
            raise RuntimeError, \
                'BUG: "node_ok" attr not deleted for %s' % node

    # Create and return a Match object.
    r = range(len(newendsnodelist))
    r.reverse()
    for i in r:
        if newendsnodelist[i].id not in newendnodesdict:
            del newendsnodelist[i]
    return Match(topstartnode, newendsnodelist)

def merge_horizontal(matches):
    """Merge a list of matches all having the same startnode.

    If your program is SLOW, here is a good place to look. Note the "walk()"
    in "merge". If you plan to merge many matches, put them in a list and
    call one of the merge fucntions once.
    """
    if isinstance(matches, Match):
        matches = (matches,)
    if len(matches) == 0:
        return EmptyMatch()

    startnode = matches[0].startnode
    endnodedict = dict()
    for match in matches:
        if not isinstance(match, Match):
            raise TypeError, 'argument is not a Match'
        if match.startnode != startnode:
            raise TypeError, 'matches must all have the same startnode'
    return merge(startnode, matches)

def merge_chain(topstartnode, matches, single=True):
    """Merge a chain of matches.

    The startnode of each match in the chain must be below an endnode of the
    previous match. If "single" is True, only the last match can have more
    than one endnode.

    If your program is SLOW, here is a good place to look. Note the walk()
    in "merge". If you plan to merge many matches, put them in a list and
    call merge once.
    """
    if isinstance(matches, Match):
        matches = (matches,)
    if len(matches) == 0:
        return EmptyMatch()

    n = len(matches)
    for i in range(n):
        if not isinstance(matches[i], Match):
            raise TypeError, 'argument is not a Match'
        if single and i < n-1 and len(matches[i].endnodes) != 1:
            raise ValueError, 'Only the last match can have more than one endnode'
        if i > 0:
            try:
                parents = \
                   matches[i].startnode.parent_chain(matches[i-1].startnode)
            except ValueError:
                raise ValueError, 'the startnode for each match must be below ' \
                       'an endnode of the previous match'
            for endnode in matches[i-1].endnodes:
                if endnode in parents:
                    break
            else:
                raise ValueError, 'the startnode for each match must be below ' \
                       'an endnode of the previous match'
    return merge(matches[0].startnode, matches)

def merge_below_node(node, matches):
    startnodes = []
    if isinstance(matches, Match):
        matches = (matches,)
    for mat in matches:
         if not isinstance(mat, Match):
             raise TypeError, 'item in "matches" not a match'
         if mat.startnode in node.children:
             startnodes.append(mat.startnode)
         else:
             raise TypeError, 'The startnode for each match must be a child or "node"'
    mat = Match(node, startnodes)
    return mat.merge_below(matches)

class SuccessfulMatch(Match):
    """Use when all one cares about a match is its success"""
    def __init__(self, startnode):
        Match.__init__(self, startnode, [])

    def __str__(self):
        return '<%s: %s>' % (self.classname, self.startnode)

    def __repr__(self):
        return '<%s: %s>' % (self.classname, repr(self.startnode))

class FailedMatch(Match):
    """Use for failed matches"""
    def __init__(self, startnode, raiseFailedMatch=NORAISE):
        Match.__init__(self, startnode, [])
        if raiseFailedMatch == RAISE:
            raise FailedMatchError, 'no match found'

    def __str__(self):
        return '<%s: %s>' % (self.classname, str(self.startnode))

    def __repr__(self):
        return '<%s: %s>' % (self.classname, repr(self.startnode))

class EmptyMatch(Match):
    """Use for a successful match with no endnodes"""
    def __init__(self, startnode):
        Match.__init__(self, startnode, [])

    def __str__(self):
        return '<%s: %s>' % (self.classname, self.startnode)

    def __repr__(self):
        return '<%s: %s>' % (self.classname, repr(self.startnode))

#>=== MatchList ==================================================

    class MatchList(list):
        def __init__(self, alist, startnode=None):
            list.__init__(self, alist)
            if isinstance(alist, MatchList):
                if startnode is not None:
                    raise TypeError, \
                     'if "alist" is a MatchList, startnode must be absent or None'
                self.startnode = alist.startnode
            else:
                if startnode is None:
                    raise TypeError, \
                     'if "alist" is not a MatchList, then "startnode" ' \
                     'must not be none'
            self.startnode = startnode
            self.checklist(self)

        def checkthis(self, match):
            if not isinstance(match, Match):
                raise TypeError, 'argument must be a Match instance'
            if match.startnode != self.startnode:
                raise ValueError, \
                  'all elements of a MatchList must have the same startnode'
            for endnode in match.endnodes:
                if endnode.parent != match.startnode:
                    raise ValueError, \
                     'all the endnodes of the matches in a MatchList ' \
                     'must be children of the startnode'

        def check(self):
            for match in self:
                self.checkthis(match)

        def checklist(self, alist):
            if isinstance(alist, MatchList):
                if self.startnode != alist.startnode:
                    raise ValueError, \
                      'all elements of a MatchList must have the same startnode'
            else:
                for match in alist:
                    self.checkthis(match)

        def __setitem__(self, index, match):
            if isinstance(index, (int, long)):
                self.checkthis(match)
                list.__setitem__(index, match)
            elif isinstance(index, slice):
                self.checklist(match)
            list.__setitem__(self, index, match)

        def append(self, match):
            length = len(self)
            self.__setitem__(slice(length, length), [match])

        def extend(self, matches):
            length = len(self)
            self.__setitem__(slice(length, length), matches)

        def insert(self, index, match):
            self.__setitem__(slice(index, index), match)

        def __add__(self, alist):
            self.checklist(alist)
            matchlist = MatchList(self)
            matchlist.extend(MatchList(alist))
            return matchlist

        def __radd__(self, alist):
            self.checklist(alist)
            matchlist = MatchList(alist)
            matchlist.extend(MatchList(self))
            return matchlist

        def __iadd__(self, alist):
            self.checklist(alist)
            self.extend(MatchList(alist))

        def __mul__(self, obj):
            raise NotImplementedError, '"__mul__" is not implemented for MatchList'

        def __rmul__(self, obj):
            raise NotImplementedError, '"__rmul__" is not implemented for MatchList'

        def __imul__(self, obj):
            raise NotImplementedError, '"__imul__" is not implemented for MatchList'
