#! /usr/bin/env python

import sys
import Node, Tree, questioner, Creators
from LN import LN, MLLN

#>=== Labels ==============================================

"""A label is a string attached to each TemplateNode. Using Python's
syntax notation, a label is
     label ::= base_label ( . num_label )?
     base_label ::= letter (letter | digit | "_")*
     num_label ::= (digit)+
Examples are "abc", "abc.123", and "A_x". The following examples are badly
formed:
     a.b.2    More than one dot.
     123      Does not start with a letter.
     _ab.34   Starts with a "_".
     abc.12d3 Only numbers are permitted after the dot. 
"""
class LabelError(Exception):
    pass

def check_label(label):
    """Checks the syntax of a label"""
    if not label[0].isalpha():
        raise LabelError, 'first part of label must start with a letter: %s'% \
             label
    if not label.isalnum():
        for c in label[1:]:
            if not c.isalnum() and not c == '_' and not c == '.':
                raise LabelError, \
                  'label must contain only letters, numerals, "_" and at most one ".": %s' % label
    if label.count('.') > 1:
        raise LabelError, 'label has more than one "."'

def parse_label(label):
    """Parses a label into its one or two parts"""
    parts = label.split('.')
    if len(parts) == 1:
        return label, None
    elif len(parts) == 2:
        return tuple(parts)
    else:
        raise LabelError, 'label has more than one "."'

# 0-99999 can be explicitly used in templates.
_UniqueNumber = 100000
def uniquify(label):
    """Attach a unique number to a label"""
    global _UniqueNumber

    _UniqueNumber += 1
    return '%s.%s' % (label, _UniqueNumber)

#>=== TemplateNode ==================================================

class TemplateNode(Node.Node):
    """Each instance of <@...@> in a template becomes a node.

    A template is stored as node.string. A label is stored as node.label.
    """
    def __init__(self, label, tree, start=None, end=None, indent=0):
        Node.Node.__init__(self, None, Node.CODE.NONE)
        self.label = label
        check_label(self.label)
        self.base_label, self.num_label = parse_label(self.label)
        self.tree = tree
        self.start = start
        self.end = end
        self.indent = indent
        self.GD = tree.GD
        self.TD = tree.TD
        self.messages = dict()
        self.text = None  # Contains the Pyrex text for the node.

    def addnode(self, parent, index='LAST'):
        """Adds self to a parent node.

        "index" can be either an int or the string "LAST". If a node is
        added, it keeps its children.
        """
        Node.Node.addnode(self, parent, index)
        # Copy the messages dict. Want to be able to change one without
        # changing the others.
        self.messages = dict(parent.messages)

    def merge_children(*args, **kwargs):
        raise NotImplementedError, '"merge_children" is intentionally unimplemented'

    def extract(*args, **kwargs):
        raise NotImplementedError, '"extract" is intentionally unimplemented'

    def kill(*args, **kwargs):
        raise NotImplementedError, '"kill" is intentionally unimplemented'

    def parents_labels(self):
        """Find the labels for the chain of parents for this node"""
        parents = self.parent_chain()
        labels = []
        for parent in parents:
            labels.append(parent.label)
        return labels

    def template_labels(self):
        """Parses templates

        Find the "<@...@>" where "..." is some name.  If "<@...@>" is
        preceeded, on a line, only by spaces, all the lines in the
        replacement text will be reindented to match "<@...@>".
        """
        counter = 0
        start = 0
        end = 0
        while 1:
            start = self.string.find('<@', end)
            if start == -1:
                return
            end = self.string.index('@>', start) + 2
            at_string = self.string[start:end]
            label = at_string[2:-2]
            if '<@' in label:
                raise ValueError, '"<@...@>" cannot be nested'

            # Is the "<@" preceded only by spaces on a line?
            i = start - 1
            while self.string[i] == ' ' and i >= 0:
                i -= 1
            if i == -1 or self.string[i] == '\n':
                indent = start - i - 1
            else:
                indent = None
            yield start, end, label, indent, counter
            counter += 1

    def get_template(self, label, num=None):
        pair = (label, num)
        if pair in self.GD.templatedict:
            return self.GD.templatedict[pair]
        return getattr(self.TD.C, label)(self)

    def children_from_template(self):
        """Recursively creates nodes in a TemplateTree"""
        # All of label or just first part?
        if self.has_children():
            raise ValueError, 'children have already been created'
        for start, end, label, indent, counter in self.template_labels():
            newnode = TemplateNode(label, self.tree, start, end, indent)
            newnode.string = newnode.get_template(label)
            newnode.addnode(self)
        for node in self.children:
            node.children_from_template()

    def text_assembler(self):
        self.text = self.string
        nodes = list(self.children)  # Want a copy.
        nodes.reverse()
        for node in nodes:
            node.text_assembler()
            self.text = '%s%s%s' % \
                 (self.text[:node.start], node.text, self.text[node.end:])

    # The simple function "reindent" and two rules makes (almost) all
    # indentation problems disappear:
    #
    #   The pieces of text generated by this program never start or end with
    #   "\n"s. But they may contain "\n"s.
    #
    #   If "<@...@>" is preceeded, on a line of the template, only by
    #   spaces, all the lines in the replacement text will be indented the
    #   same as "<@...@>".
    #
    # If a line in a template consists of zero or more spaces followed by a
    # "<@...@>" and the "<@...@>" is replaced by "", then there will be an
    # an unnecessary line in the output.
    def reindent(self, text, indent):
        """Reindent a piece of text"""
        lines = text.splitlines()
        sep = '\n' + ' '*indent
        return sep.join(lines)

    # The next three functions are used to send information down the node
    # tree. A node labeled "abc" can read messages addressed to "abc". The
    # node labeled "abc.123" can read messages addresed to "abc" or "abc.123".
    def send_message(self, label, obj):
        """Send a message down the node tree"""
        self.messages[label] = obj

    def get_message(self):
        """Get a message.

        A uniqueified Node "xxx.nnn" can access information sent to "xxx" but
        not vice versa.
        """
        id, num = parse_label(self.label)
        if not self.label in self.messages and \
                 not self.base_label in self.messages:
            sys.stdout.flush()
            raise KeyError, 'No message found for %s in\n%s\nParents are %s' % \
                    (self.label, self.messages, self.parents_labels())
        if self.label in self.messages:
            return self.messages[self.label]
        else:
            return self.messages[self.base_label]

    def del_message(self):
        """Delete a message"""
        del self.messages[self.label]

    # A tree-message is used to send something to some specific node. The
    # receiver node must have a 2-part name (abcd.123). The object which is
    # sent is a dictionary of messages which is merged with the messages
    # dictionary of the receiving node. If "override" is True, these
    # messages override messages in the receiving dictionary.
    def send_tree_message(self, label, message_dict, override=True):

        """A tree-message is used to send a message to some specific node.

        The receiver node must have a 2-part name (abcd.123).
        """
        check_label(label)
        if label.count('.') == 0:
            raise LabelError, \
             'a node whose label is without a "." cannot process tree-messages: %s' % \
                 self.label
        if label in self.TD.tree_messages:
            raise KeyError, 'tree-message to "%s" has already been sent from "%s"' % \
                   (label, self.label)
        if not isinstance(message_dict, dict):
            raise TypeError, 'message must be a dictionary'
        message_dict[None] = override
        self.TD.tree_messages[label] = dict(message_dict)

    def get_tree_message(self):
        """Get a tree-message"""
        if self.label.count('.') != 1:
            raise KeyError, \
             'a node whose label is without a "." cannot process tree messages: %s' % \
                 self.label
        if not self.label in self.tree.tree_messages:
            sys.stdout.flush()
            raise KeyError, \
               'No tree-message found for %s in\n%s\nParents are %s' % \
                    (self.label, self, self.parents_labels())
        message_dict = self.tree.tree_messages[self.label]
        if message_dict[None]:  # override
            self.messages.update(message_dict)
        else:
            for key in message_dict:
                if key not in self.messages:
                     self.messages[key] = message_dict[key]
        del self.TD.tree_messages[self.label]

    def del_tree_message(self):
        """Delete a tree-message"""
        if self.label.count('.') != 1:
            raise KeyError, \
             'a node whose label is without a "." cannot process tree_messages: %s' % \
                 self.label
        del self.TD.tree_messages[self.label]

#>=== TemplateTree ================================================

class _TreeData(object):
    def __init__(self, Q, C):
        self.Q = Q
        self.C = C
        # The keys of "tree_messages" are dotted labels. Each item is a
        # dictionary.
        self.tree_messages = dict()

class TemplateTree(Tree.Tree):
    def __init__(self, parsetree, GD):
        self.id = Tree.Tree.counter
        Tree.Tree.counter += 1
        self.tokenstrings = parsetree.get_tokens()
        self.src = ' '.join(self.tokenstrings)
        self.GD = GD  # GlobalData.
        Q = questioner.Questioner(parsetree)
        C = Creators.Creators(Q)
        self.TD = _TreeData(Q, C)
        top = TemplateNode('top', self)
        self.addnode(top, 'TOP')
        self.top.string = self.top.get_template('top')

    def strip_const_semi(self):
        out = []
        for string in self.tokenstrings:
            if string not in ('const', ';'):
                out.append(string)
        return out

    def addnode(self, node, parent, index='LAST'):
        Tree.Tree.addnode(self, node, parent, index)
        # Copy the messages dict. Want to be able to change one without
        # changing the others.
        if parent == 'TOP':
            node.messages = dict()
        else:
            node.messages = dict(parent.messages)

    def copy(self, tree):
        raise NotImplementedError, '"copy" is intentionally unimplemented'

    def stripNODENAMEleaves(self):
        raise NotImplementedError, '"stripNODENAMEleaves" is intentionally unimplemented'

if __name__ == '__main__':
    strings = ('A', 'B')
    text = '1A2B3A4'
    print split_before_many(text, strings)
