#! /usr/bin/env python

import sys
from LN import LN, MLLN
import Node, Match

#> === Matches along a chain of nodes ============================

class BaseChainPattern(tuple):
    def __new__(typ, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]
        for obj in args:
            if not isinstance(obj, (str, BaseChainPattern)):
                raise TypeError, \
                 'argument is not a string, ChainPattern, OR_SKIP, or ANY. ' \
                 'It is a %s.' % obj.__class__.__name__
        return tuple.__new__(typ, args)

    def __init__(self, *args, **kwargs):
        if 'speller' in kwargs:
            nodestringdict = kwargs['speller']
            self.spelling_check(nodestringdict)

    def __str__(self):
        argstrs = []
        for arg in tuple(self):
            if isinstance(arg, str):
                argstrs.append('"%s"' % arg)
            else:
                argstrs.append(self.classname + ' obj')
        middle = ', '.join(argstrs)
        return '%s(%s)' % (self.classname, middle)

    def __repr__(self):
        return '%s( %s )' % (self.classname, tuple.__repr__(self))

    def spelling_check(self, nodestringdict):
        for obj in self:
            if isinstance(obj, str) and obj not in nodestringdict:
                raise ValueError, '"%s" is spelled wrong' % obj
            elif isinstance(obj, BaseChainPattern):
                obj.spelling_check()

    def get_classname(self):
        return self.__class__.__name__
    classname = property(get_classname)

class ChainPattern(BaseChainPattern):
    """Create an instance by:
         ChainPattern(atuple, adict)
         ChainPattern(atuple)
         ChainPattern(patt0, ..., pattn, adict)
         ChainPattern(patt0, ..., pattn)
            where "atuple" is a tuple of strings or ChainPatterns,
            patt0, ..., pattn are strings or ChainPatterns, and "adict"
            contains all the strings that are permitted in the pattern. 
            "adict" defaults to None. In this case spelling checks are not
            done.
    """
    def check(self, node, n=0):
        if n >= len(self):
            return False
        obj = self[n]
        if isinstance(obj, str):
            return obj == node.string
        return obj.check(node)            

    def _walk(self, startnode, startdepth):
        localdepth = startnode.depth - startdepth
        ck = self.check(startnode, localdepth)
        if localdepth == len(self) - 1 and ck:
            # Successful match.
            yield startnode
        if localdepth >= len(self) or not ck:
            # Failed match.
            return
        for node in startnode.children:
            for node2 in self._walk(node, startdepth):
                yield node2

    def matcher(self, startnode):
        # If there is no match, "matcher" will exit without yielding
        # anything.
        if startnode.string != self[0]:
            raise ValueError, 'startnode does not match pattern:\n%s\n%s' % \
                (startnode, self) 
        for endnode in self._walk(startnode, startnode.depth):
            yield Match.Match(startnode, endnode)

    def all_matches(self, startnode):
        out = []
        for match in self.matcher(startnode):
            out.append(match)
        return out

    def all_matches_as_one_match(self, startnode):
        matchlist = self.all_matches(startnode)
        if len(matchlist) == 0:
            return Match.EmptyMatch(startnode)
        for match in matchlist:
            if startnode != match.startnode:
                raise TypeError, 'all matches must have same startnode'
        return Match.merge(startnode, matchlist)

    def first_match(self, startnode, raiseFailedMatch=False):
        # If there are no matches, the following trick sets "mat" to "None".
        mat = None
        for mat in self.matcher(startnode):
            break
        if mat is None:
            return Match.FailedMatch(startnode, raiseFailedMatch)
        return mat

# Not used much so far. Uses first_match approach.
class OR_SKIP(BaseChainPattern):
    """An OR_SKIP pattern matches a node if any of the strings in it match the node.

    When searching a list of children for an OR_SKIP match, the first match
    found is returned. Nodes preceeding the first match are skipped.
    """
    def check(self, node):
        if node.string in self:
            return True
        return False

class ANY(BaseChainPattern):
    """Matches anything"""
    def __new__(typ):
        return tuple.__new__(typ, ())  # Empty tuple.

    def __str__(self):
        return 'ANY()'

    def __repr__(self):
        return 'ANY()'

    def check(self, node):
        return True

class CTOKEN(BaseChainPattern):
    """Matches anything"""
    def __new__(typ):
        return tuple.__new__(typ, ())  # Empty tuple.

    def __str__(self):
        return 'CTOKEN()'

    def __repr__(self):
        return 'CTOKEN()'

    def check(self, node):
        return node.code == Node.CODE.CTOKEN

#> === Patterns for matches within a set of children ===================

class ChildPattern(tuple):
    def __new__(typ, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]
        for obj in args:
            if not isinstance(obj, (str, ChildPattern)):
                raise TypeError, \
                 'argument sequence must contain strings or ChildPatterns only'
        return tuple.__new__(typ, args)

    def __init__(self, *args, **kwargs):
        """If the keyword argument speller=tree.speller is present, check
        the spelling of all the strings in the pattern, no matter how deeply
        nested. The keys of tree.speller are all the node strings for tree.
        """
        if 'speller' in kwargs and kwargs['speller'] is not None:
            nodestringdict = kwargs['speller']
            self.spelling_check(nodestringdict)

    def __str__(self):
        argstrs = []
        for arg in tuple(self):
            if isinstance(arg, str):
                argstrs.append('"%s"' % arg)
            else:
                argstrs.append(self.classname + ' obj')
        middle = ', '.join(argstrs)
        return '%s(%s)' % (self.classname, middle)

    def __repr__(self):
        return '%s( %s )' % (self.classname, tuple.__repr__(self))

    def spelling_check(self, nodestringdict):
        for obj in self:
            if isinstance(obj, str) and obj not in nodestringdict:
                raise ValueError, '"%s" is spelled wrong' % obj
            elif isinstance(obj, ChildPattern):
                obj.spelling_check(nodestringdict)

    def get_classname(self):
        return self.__class__.__name__
    classname = property(get_classname)

    def one_past_end(self, parent, start):
        if start < 0 or start > len(parent.children):  # Note the ">".
            return IndexError, '"start" out of range'
        if start == len(parent.children):
            return True
        return False

    def check_node(self, patt_index, parent, start):
        """Matches a pattern in self against the children of parent.

        "pattern_index" chooses a pattern in "self".
        "parent" is a node.
        "start" is the index of one of the children of "parent".
        A Match is returned whose startnode is "parent". "check_node" and 
        "match" call each other recursively.
        """
        patt = self[patt_index]
        children = parent.children
        if self.one_past_end(parent, start):
            return start, Match.FailedMatch(parent)
        child = children[start]
        if isinstance(patt, str):
            if child.string == patt:
                return start + 1, Match.Match(parent, child)
            return start, Match.FailedMatch(parent)
        return patt.match(parent, start)

    def match(self, parent, start, which_endnode=None):
        """Returns a Match between the current pattern and parent.children[start:]

        This function must be overriden by each subclass.
        """
        raise NotImplementedError, 'ChildPattern.match must be overridden'

class Next(ChildPattern):
    # When I get around to it, this will be part of an implementation of
    # "matcher" using ChildPattern. Undebugged.
    def __init__(self, patt, **kwargs):
        ChildPattern.__init__(self, patt, **kwargs)
        if len(self) != 1:
            raise TypeError, '"ZeroOrMore" takes exactly one argument'

    def match(self, parent, start):
        if self.one_past_end(parent, start):
            return start, Match.EmptyMatch(parent)
        start1 = start
        while 1:
            start1, match1 = self.check_node(0, parent, start1)
            if match1.success:
                return start1, match1
            if start1 >= len(parent.children):
                return start1, Match.EmptyMatch(parent)

class ZeroOrOne(ChildPattern):
    def __init__(self, patt, **kwargs):
        ChildPattern.__init__(self, patt, **kwargs)
        if len(self) != 1:
            raise TypeError, '"ZeroOrOne" takes exactly one argument'

    def match(self, parent, start):
        if self.one_past_end(parent, start):
            newstart, newmatch = start, Match.EmptyMatch(parent)
        start1, match1 = self.check_node(0, parent, start)
        if match1.failure or start1 >= len(parent.children):
            return start, Match.EmptyMatch(parent)
        return start1, match1

class ZeroOrMore(ChildPattern):
    def __init__(self, patt, **kwargs):
        ChildPattern.__init__(self, patt, **kwargs)
        if len(self) != 1:
            raise TypeError, '"ZeroOrMore" takes exactly one argument'

    def match(self, parent, start):
        if self.one_past_end(parent, start):
            return start, Match.EmptyMatch(parent)
        start1 = start
        matches = list()
        while 1:
            start1, match1 = self.check_node(0, parent, start1)
            if match1.failure:
                return match1
            matches.append(match1)
            if start1 >= len(parent.children):
                break
        if len(matches) == 0:
            return start1, Match.EmptyMatch(parent)
        return start1, Match.merge_horizontal(matches)

class OneOrMore(ChildPattern):
    def __init__(self, patt, **kwargs):
        ChildPattern.__init__(self, patt, **kwargs)
        if len(self) != 1:
            raise TypeError, '"OneOrMore" takes exactly one argument'

    def match(self, parent, start):
        if self.one_past_end(parent, start):
            return start, Match.FailedMatch(parent)
        start1 = start
        matches = list()
        while 1:
            start1, match1 = self.check_node(0, parent, start1)
            if match1.failure:
                break
            matches.append(match1)
            if start1 >= len(parent.children):
                break
        if len(matches) == 0:
            return start1, Match.FailedMatch(parent)
        return start1, Match.merge_horizontal(matches)

class OR(ChildPattern):
    def __init__(self, *args, **kwargs):
        ChildPattern.__init__(self, *args, **kwargs)
        if len(self) < 2:
            raise TypeError, '"OR" takes at least one argument'

    def match(self, parent, start):
        if self.one_past_end(parent, start):
            return start, Match.FailedMatch(parent)
        for i in range(len(self)):
            start1, match1 = self.check_node(i, parent, start)
            if match1.success:
                return start1, match1
        return start1, Match.FailedMatch(parent)

class ADJ(ChildPattern):
    def __init__(self, *args, **kwargs):
        ChildPattern.__init__(self, *args, **kwargs)
        if len(self) == 0:
            raise TypeError, '"ADJ" takes at least one argument'

    def match(self, parent, start):
        start1 = start
        matches = list()
        startnodes = list()
        for i in range(len(self)):
            start1, match1 = self.check_node(i, parent, start1)
            if match1.failure:
                return start1, Match.FailedMatch(parent)
            matches.append(match1)
            startnodes.append(match1.startnode)
        if len(matches) == 0:
            return start1, Match.EmptyMatch(parent)
        return start1, Match.merge_horizontal(matches)

class NTH(ChildPattern):
    def match(self, parent, start):
        if start < 0 or start >= len(parent.children):
            raise IndexError, 'index out of range'
        return self.check_node(0, parent, start)

if __name__ == '__main__':
    from TokenTree import TokenTree as _TokenTree

    def doit1(tree, patt_tup):
        patt = ChainPattern(patt_tup, speller=tree.speller)
        LN(patt)
        print '************'
        for match in patt.matcher(tree.top):
            LN(match)
        print '*************'
        LN(patt.first_match(tree.top, Match.NORAISE))
        print '*************'

    def doit2(tree, patt_tup):
        patt = ChainPattern(patt_tup, speller=tree.speller)
        match = patt.first_match(tree.top, Match.RAISE)
        startnode = match.the_endnode()
        LN(startnode)
        LN(startnode.children[0])
        childpatt = ZeroOrOne('storageClassSpecifier', speller=tree.speller)
        start, out = childpatt.match(startnode, 0)
        LN(start, out)
        childpatt2 = OneOrMore('typeSpecifier', speller=tree.speller)
        start2, out2 = childpatt2.match(startnode, start)
        LN(start2, out2)

    def doit3(tree):
        patt_tup = ['externalList', 'externalDef', 'declaration', 'declSpecifiers']
        patt = ChainPattern(patt_tup, speller=tree.speller)
        pattmatch = patt.first_match(tree.top, Match.RAISE)
        startnode = pattmatch.the_endnode()
        LN(startnode)
        LN(startnode.children[0])
        childpatt1 = ZeroOrOne('storageClassSpecifier')
        childpatt2 = OneOrMore('typeSpecifier')
        childpatt = ADJ(childpatt1, childpatt2, speller=tree.speller)
        start, out = childpatt.match(startnode, 0)
        LN(start)
        LN(out)

    def doit3X(tree):
        """Put a "X" in the middle of any string below and run"""
        patt_tup = ['externalList', 'externalDef', 'declaration', 'declSpecifiers']
        patt = ChainPattern(patt_tup, speller=tree.speller)
        pattmatch = patt.first_match(tree.top, Match.RAISE)
        startnode = pattmatch.the_endnode()
        LN(startnode)
        LN(startnode.children[0])
        ZeroOrMore('stXorageClassSpecifier', speller=tree.speller)
        OR('storageClassSpecifier', speller=tree.speller)
        childpatt1 = ZeroOrOne('stXorageClassSpecifier')
        childpatt2 = OneOrMore('typeSpecifier')
        childpatt = ADJ(childpatt1, childpatt2, speller=tree.speller)
        start, out = childpatt.match(startnode, 0)
        LN(start)
        LN(out)

    tree = _TokenTree(open('pico.h.data', 'r').read())
    print tree
    list1 = ['externalList', 'externalDef', 'declaration',
                   'declSpecifiers', 'typeSpecifier',
                    ANY(), OR_SKIP('unsigned', 'long')]
    list2 = ['externalList', 'externalDef', 'declaration',
                   'declSpecifiers', 'typeSpecifier',
                   'typeSpecifier.4', 'long']
    list3 = ['externalList', 'externalDef', 'declaration', 'declSpecifiers']
    list3X = ['externalList', 'externalDef', 'declaration', 'declSpecXifiers']
    list4 = ['externalList', 'xxx']
    list5 = ['externalList', 'externalDef']
#    doit1(tree, list1)
#    doit2(tree, list3)
    doit3X(tree)
#    patt = Pattern('externalList', 'externalDef')
#    for match in patt.matcher(tree.top):
#        LN(match.endnode.print_as_tree())
