#! /usr/bin/env python

import sys
import globals
import Qstruct
from Match import RAISE, NORAISE, FailedMatchError, \
     Match, FailedMatch, EmptyMatch, merge_horizontal, merge_chain, \
     merge_below_node
from pattern import ChainPattern, OR_SKIP, ANY, \
     ZeroOrOne, ZeroOrMore, OneOrMore, OR, ADJ
from TokenTree import TokenTree
from LN import LN, MLLN

class Questioner(Qstruct.Qstruct):
    def __init__(self, tree):
        self.tree = tree
        self.nodestringdict = dict()
        for node in self.tree.walk():
            self.nodestringdict[node.string] = None

    def AllMatches(self, patt_tuple, startnode):
        patt = ChainPattern(patt_tuple)
        if startnode is None:
            startnode = self.tree.top
        return patt.all_matches(startnode)

    def FirstMatch(self, patt_tuple, startnode, raiseFailedMatch=NORAISE):
        patt = ChainPattern(patt_tuple)
        return patt.first_match(startnode, raiseFailedMatch)

    def YieldMatches(self, patt_tuple, startnode,  raiseFailedMatch=NORAISE):
        patt = ChainPattern(patt_tuple)
        if startnode is None:
            startnode = self.tree.top
        for match in patt.matcher(startnode, raiseFailedMatch):
            yield match

    def ExternalDef(self):
        """Returns a list of all the externalDef's as new trees."""
        # These functions look all the same after a while. But they 
        # return different things. State what is being returned in the
        # doc line.
        patt_tup = ('externalList', 'externalDef')
        matches = self.AllMatches(patt_tup, self.tree.top)
        trees = []
        for match in matches:
            node = match.the_endnode()
            node.extract()
            newtree = TokenTree(node)
            trees.append(newtree)
        return trees

    def Declarator(self):
        """Finds the "declarator" node starting at top of tree"""
        patt_tup = ('externalDef', 'declaration',
                     'initDeclList', 'initDecl', 'declarator')
        return self.FirstMatch(patt_tup, self.tree.top)

    def IsDeclarator_0_3(self, node):
        """Are the first two children of "node" a declarator.0 and a
        declarator.3?"""
        child_patt = ADJ(ZeroOrOne('pointerGroup'), 'declarator.0',
                             'declarator.3')
        nextindex, mat = child_patt.match(node, 0)
        return mat.success

    def IsPrototype(self):
        match = self.Declarator()
        if match.failure:
            return False
        return self.IsDeclarator_0_3(match.the_endnode())

    def IsFunctionPointer(self):
        match1 = self.Declarator()
        if match1.failure:
            return False
        node1 = match1.the_endnode()
        child_patt = ADJ(ZeroOrOne('pointerGroup'), 'declarator.1',
                             'declarator')
        nextindex, mat = child_patt.match(node1, 0)
        if mat.failure:
             return False
        node2 = mat.endnodes[-1]
        if node2.children[0].string == 'pointerGroup':
            return True
        return False

    def DeclSpecifiers(self):
        patt_tup = ('externalDef', 'declaration', 'declSpecifiers')
        return self.FirstMatch(patt_tup, self.tree.top)

    def Typedef(self, startnode):
        patt_tup0 = ('storageClassSpecifier', 'storageClassSpecifier.2',
               'typedef')
        if len(startnode.children) == 0:
            return FailedMatch(patt_tup0, startnode)
        # We don't know what "startnode.children[0].string" is. If it is
        # not "storageClassSpecifier", we want a FailedMatch returned.
        if startnode.children[0].string != 'storageClassSpecifier':
            return FailedMatch(startnode)
        fm = self.FirstMatch(patt_tup0, startnode.children[0])
        return merge_below_node(startnode, fm)

    def TypedefStructEnumOrUnion(self):
        match1 = self.DeclSpecifiers()
        if match1.failure:
            return match1
        match2 = self.Typedef(match1.the_endnode())
        if match2.failure:
            return match2
        node = match1.the_endnode()
        if len(node.children) < 2 or \
                  node.children[1].string != 'typeSpecifier':
            # Eventually use a match tree instead of None here.
            return FailedMatch(None, self.tree.top)
        patt_tup1 = ('typeSpecifier',
            OR_SKIP('structSpecifier', 'enumSpecifier', 'unionSpecifier'),
            OR_SKIP('structSpecifier.0', 'enumSpecifier.0', 'unionSpecifier.0'),
            OR_SKIP('struct', 'enum', 'union'))
        return self.FirstMatch(patt_tup1, node.children[1])

    def IsTypedefStructEnumOrUnion(self):
        return self.TypedefStructEnumOrUnion().success

    def IsTypedefStruct(self):
        match = self.TypedefStructEnumOrUnion()
        if match.failure:
            return False
        return match.the_endnode().string == 'struct'

    def IsTypedefEnum(self):
        match = self.TypedefStructEnumOrUnion()
        if match.failure:
            return False
        return match.the_endnode().string == 'enum'

    def IsTypedefUnion(self):
        match = self.TypedefStructEnumOrUnion()
        if match.failure:
            return False
        return match.the_endnode().string == 'union'

    def TypedefSimple(self):
        match1 = self.DeclSpecifiers()
        if match1.failure:
            return match1
        match2 = self.Typedef(match1.the_endnode())
        if match2.failure:
            return FailedMatch(self.tree.top)
        node = match1.the_endnode()
        if len(node.children) <= 1:
            return merge_chain(self.tree.top, (match1, match2) )
        patt_tup1 = ('typeSpecifier', 'structSpecifier',
               'structSpecifier.0', 'struct')
        match3 = self.FirstMatch(patt_tup1, node.children[1])
        if match3.failure:
            return merge_chain(self.tree.top, (match1, match2) )
        return FailedMatch(self.tree.top)

    def IsTypedefSimple(self):
        return self.TypedefSimple().success

    def IsTypedef(self):
        declSpecMatch = self.DeclSpecifiers()
        if declSpecMatch.failure:
            return False
        patt_tup = ('externalDef', 'declaration', 'declSpecifiers',
               'storageClassSpecifier', 'storageClassSpecifier.2',
               'typedef')
        return self.FirstMatch(patt_tup, self.tree.top)

    def StructSpecifier(self):
        patt_tup = ('externalDef', 'declaration', 'declSpecifiers',
            'typeSpecifier', 'structSpecifier')
        return self.FirstMatch(patt_tup, self.tree.top)

    def BasicStruct(self):
        match = self.StructSpecifier()
        if match.failure:
            return match
        node = match.the_endnode()
        child_patt = ADJ('structSpecifier.0', 'structOrUnionBody')
        index, match1 = child_patt.match(node, 0)
        if match1.failure:
            return FailedMatch(match.startnode)
        return merge_chain(match.startnode, (match, match1))

    def ForwardDeclaration(self, BasicStructmatch):
        if BasicStructmatch.failure:
            return BasicStructmatch
        StructOrUnionBodyNode = BasicStructmatch.endnodes[1]
        child_patt = ADJ('structOrUnionBody.4')
        index, match = child_patt.match(StructOrUnionBodyNode, 0)
        if match.success:
            tokenstring = match.endnodes[0].children[0].string
            globals.GlobalData.forward_declarations.append(tokenstring)
        return match

    def StructDeclarationList(self):
        patt_tup = ('externalDef', 'declaration', 'declSpecifiers',
            'typeSpecifier', 'structSpecifier', 'structOrUnionBody',
            'structDeclarationList')
        return self.FirstMatch(patt_tup, self.tree.top)

    def TypeSpecifier(self):
        ORpatt = OR_SKIP(
            'typeSpecifier.0', 'typeSpecifier.1', 'typeSpecifier.2',
            'typeSpecifier.3', 'typeSpecifier.4', 'typeSpecifier.5',
            'typeSpecifier.6', 'typeSpecifier.7', 'typeSpecifier.8')  

        patt_tup = ('externalDef', 'declaration', 'declSpecifiers',
            'typeSpecifier', ORpatt)
        return self.FirstMatch(patt_tup, self.tree.top)
