#! /usr/bin/env python

import os, sys
from myexceptions import *
from LN import LN, MLLN

"""Two ways of designing a tree:
1. A set of nodes with a special top node.
2. A distinct object pointing to a top node.
For (1), each node is the top node of a tree. This makes it easier to
manipulate subtrees. But then I need two __repr__'s, one to print a node and
one to print the tree headed by the node. But worse, all of TokenTree would
have to be put in Token making Token too big. I decided to put some of the
tree methods in Node so I can both manipulate subtrees and keep TokenTree
separate. Ugly. It seems to work.
"""

#>=== class Node =================================================

class CODE(object):
    NONE, LP, RP, NODENAME, CTOKEN, LEAF = range(6)
    names = ('NONE', 'LP', 'RP', 'NODENAME', 'CTOKEN', 'LEAF')

class WALKAROUND_CODE(object):
    FIRST, SECOND, LEAF = 0, 1, 2

class Node(object):
    counter = 0

    def __init__(self, string=None, code=None):
        self.id = Node.counter
        Node.counter += 1
        self.parent = None
        self.children = list()
        self.left = None
        self.right = None
        self.top = self
        self.string = string
        if code is None:
            code = CODE.NONE
        self.code = code
        self.depth = 0
        self.index = None
        self.tree = None

    # Create ".classname" attribute for this class and all its descendents.
    def get_classname(self):
        return self.__class__.__name__
    classname = property(get_classname)

    def copy(self):
        """Copies a node."""
        # Override me.
        n = Node()
        return n

    def set_depths(self, depth=0):
        self.depth = depth
        for child in self.children:
            child.set_depths(depth+1)

    def which_child_am_I(self):
        """Finds index of a Node in the list of its parent's children."""
        if self.parent is None:
            raise ValueError, 'node has no parent'
        i = self.parent.children.index(self)
        return i

    def set_index(self, i=None):
        if i is None:
            if self.parent is None:
                self.index = None
            else:
                self.index = self.which_child_am_I()
        else:
            self.index = i

    def set_child_indices(self):
        for i, node in enumerate(self.children):
            node.index = i

    def has_children(self):
        return len(self.children) > 0

    def __str__(self):
        return '<%s: %i, |%s|, %s>' % \
              (self.classname, self.id, self.string,
                    CODE.names[self.code])

    def print_as_tree(self):
        out = []
        for node in self.walk():
            s = '%s%s' % (node.depth * '  ', node)
            out.append(s)
        return '\n'.join(out)

    def __repr__(self):
        if self.parent is None:
            p = 'None'
        else:
            p = self.parent.id
        if self.left is None:
            ls = 'None'
        else:
            ls = self.left.id
        if self.right is None:
            rs = 'None'
        else:
            rs = self.right.id
        if self.top is None:
            to = 'None'
        else:
            to = self.top.id
        if self.tree is None:
            tr = 'None'
        else:
            tr = id(self.tree)
        text = '<%s %i, |%s|, %s\n' \
               '    parent: %s, left: %s, right: %s\n' \
               '    depth: %s, index: %s,\n' \
               '    top: %s, tree: %s>' % \
             (self.classname, self.id, self.string, CODE.names[self.code],
              p, ls, rs,
              self.depth, self.index,
              to, tr)
        return text

    def partial_reset(self):
        """Reinitializes a Node except for node.code and node.string"""
        self.parent = None
        self.left = None
        self.right = None
        self.children = list()
        self.top = self
        self.depth = None
        self.index = None
        self.tree = None

    def reset(self):
        """Reinitializes a Node."""
        self.partial_reset()
        self.string = str(self.id)
        self.code = CODE.NONE

    def parent_chain(self, node=None):
        """Find the chain of parents for a node.

        Find the chain of parents up to and including node. "self"
        is included in the chain. If "node" is None, the chain goes to
        the top of its tree.
        """
        parents = []
        node_found = False
        if node is None:
            node_found = True
        n = self
        while n is not None:
            parents.append(n)
            if n == node:
                node_found = True
                break
            n = n.parent
        if not node_found:
            print parents
            raise ValueError, '"node" not in parent chain'
        return parents

    def is_descendent_of(self, node):
        """Is "node" an ancestor of "self"?"""
        if self == node or self.parent is None:
            return False
        n = self
        while n is not None:
            if n == node:
                return True
            n = n.parent
        return False

    def asstring(self):
        """Simplest string describing a Node."""
        # Override me.
        return str(self.id)

    def isleaf(self):
        """Is this Node a leaf?"""
        return len(self.children) == 0

    def leaves(self):
        out = []
        for node in self.walk():
            if len(self.children) == 0:
                out.append(node)
        return out

    def walk(self):
        """Generator that does top-down, left-right traversal for node and below."""
        yield self
        for child in self.children:
            for node in child.walk():
                yield node

    def reverse_walk(self):
        """Generator that does bottom-up, right-left traversal below node.

        Use this if you plan to delete nodes.
        """
        r = len(self.children)
        for i in range(r-1, -1, -1):
            for node in self.children[i].reverse_walk():
                yield node
        yield self

    def walkaroundNode(self):
        """Walks around the outside of the tree structure headed by node.

        Each Node that has children is visited twice. The first visit is
        before its children are visited. The second visit is after its
        children are visited. Use the WALKAROUND codes to detect which visit
        is being done. Leaves are visited only once.

        Do not have more than one walkaround iterator at a time for a tree.
        Otherwise the codes will be messed up. I should rewrite this so that
        the codes are kept in a weakref dict inside this function.
        """
        walkaround_codes = dict()
        if len(self.children) > 0:
            walkaround_codes[self.id] = WALKAROUND_CODE.FIRST
            yield walkaround_codes[self.id], self
            for child in self.children:
                for code, node in child.walkaroundNode():
                    yield code, node
            walkaround_codes[self.id] = WALKAROUND_CODE.SECOND
            yield walkaround_codes[self.id], self
            del walkaround_code[self.id]
        else:
            walkaround_codes[self.id] = WALKAROUND_CODE.LEAF
            yield walkaround_codes[self.id], self
            del walkaround_codes[self.id]

    def walk_parents(self, n=0):
        """Walks, starting from a node, yielding the nodes and a list of their parents.

        Skip the first "n" nodes. The parent list includes the node itself.
        """
        parents = []
        counter = 0
        for node in self.walk():
            # Bypass nodes if counter is < n.
            if counter == n:
                parents = self.parent_chain()
            elif counter > n:
                parents = parents[:node.depth] + [node]
                yield node, parents
            counter += 1

    def addnode(self, parent, index='LAST'):
        """Adds self to a parent node.

        "index" can be either an int or the string "LAST". If a node is
        added, it keeps its children.
        """
        if not isinstance(parent, Node):
            raise TypeError, '"parent" must be a Node'
        if index == 'LAST':
            index = len(parent.children)
        if index < 0 or index > len(parent.children):  # Note the ">".
            raise IndexError, '"index" out of range'

        self.parent = parent

        # Set the left sibling.
        if index == 0 or len(parent.children) == 0:
            self.left = None
        else:
            self.left = parent.children[index-1]
            self.left.right = self

        # Set the right sibling.
        if index == len(parent.children) or len(parent.children) == 0:
            self.right = None
        else:
            self.right = parent.children[index]
            self.right.left = self

        parent.children.insert(index, self)
        self.set_depths(parent.depth + 1)
        parent.set_child_indices()
        self.tree = parent.tree
        for node in self.walk():
            node.tree = parent.tree
            node.top = parent.top

    def merge_children(self, newparent, index='LAST'):
        """Add the children of a node to newparent.

        "index" can be either an int or the string "LAST". The node is
        partially reset.
        """
        if not isinstance(newparent, Node):
            raise TypeError, '"newparent" must be a Node'
        if index == 'LAST':
            index = len(newparent.children)
        if not isinstance(index, int):
            raise TypeError, '"index" must be an int or the string "LAST"'
        if index < 0 or index > len(newparent.children):
            raise IndexError, '"index" out of range'
        if len(self.children) == 0:
            return
        for node in self.children:
            node.parent = newparent

        # Set the left sibling.
        if index == 0 or len(newparent.children) == 0:
            self.children[0].left = None
        else:
            self.children[0].left = newparent.children[index-1]
            self.children[0].left.right = self.children[0]

        # Set the right sibling.
        if index == len(newparent.children) or len(newparent.children) == 0:
            self.children[-1].right = None
        else:
            self.children[-1].right = newparent.children[index]
            self.children[-1].right.left = self.children[-1]

        for node in self.children:
            node.set_depths(newparent.depth + 1)
            for node2 in node.walk():
                node2.top = newparent.top
                node2.tree = newparent.tree
        newparent.children[index:index] = self.children
        newparent.set_child_indices()

        self.partial_reset()

    def extract(self):
        """Separate a Node from its parent. Its children go with it."""
        if self.parent is None:
            raise TypeError, 'node has no parent'

        if self.left is not None:
            self.left.right = self.right
            self.left = None
        if self.right is not None:
            self.right.left = self.left
            self.right = None

        del self.parent.children[self.index]
        self.set_depths()
        self.parent.set_child_indices()
        self.parent = None
        self.index = None
        for node in self.walk():
            node.tree = None
            node.top = self

    def kill(self):
        """Removes a node from a tree. It is replaced by its children."""
        if self.parent is None:
            raise TypeError, 'node has no parent'
        if self.isleaf():
            self.extract()
        else:
            parent = self.parent
            index = self.index
            children = self.children
            self.extract()
            self.merge_children(parent, index)
