from __future__ import generators

import sys, time
from linenum import LN
from externalDef import DEBUG

# Used in "CREATORs"
import treematch
from treematch import NoMatchException

class LabelError(Exception):
    pass

def check_label(label):
    """Checks the syntax of a label"""
    if not label[0].isalpha():
        raise LabelError, 'first part of label must start with a letter: %s'% \
             label
    if not label.isalnum():
        for c in label[1:]:
            if not c.isalnum() and not c == '_' and not c == '.':
                raise LabelError, \
                  'label must contain only letters, numerals, "_" and at most one ".": %s' % label
    if label.count('.') > 1:
        raise LabelError, 'label has more than one "."'

def parse_label(label):
    """Parses a label into its one or two parts"""
    parts = label.split('.')
    if len(parts) == 1:
        return label, None
    elif len(parts) == 2:
        return tuple(parts)
    else:
        raise LabelError, 'label has more than one "."'

# 0-999999 can be explicitly used in templates.
def unique_number():
    """Generator returning the natural numbers >= 1000000"""
    i = 1000000
    while 1:
        yield i
        i += 1

# Create a unique label. Used to send messages to a specific node.
UniNum = unique_number()
def make_unique_label(label):
    return '%s.%s' % (label, UniNum.next())

class All:
    """Global information"""
    def __init__(self, template_data, labels, metanodes):
        self.template_data = template_data
        self.labels = labels
        self.metanodes = metanodes
        # These keep track of incomplete declarations.
        self.finished_structs = []
        self.forward_structs = dict()

class Tree:
    """supervises a hierarchy of Nodes"""
    def __init__(self, top_label, which, treedata, all):
        self.top_label = top_label
        self.template_data = all.template_data
        self.labels = all.labels
        self.which = which
        self.treedata = treedata
        self.all = all
        self.global_messages = dict()
        self.topnode = None
        self.nodes = dict()
        self.metanodes = all.metanodes
        self.nodecounter = 0

    def createNodes(self):
        """Recursively creates all the Nodes"""
        self.topnode = Node(self.top_label, self, None)

# Information from the templates file (CREATORs) is put here. Each Node has
# an associated MetaNode.
class MetaNode:
    """A MetaNode is associated with each entry in the template file"""

    stub_template = """\
def outer():
    def Creator_Stub(node):
        return '<STUB: %s>' % node.label
    return Creator_Stub
"""

    def __init__(self, label, template_data):
        self.label = label
        self.global_messages = {}
        # Putting the expanders in Node may make debugging easier. Therefore
        # MetaNode only knows their names.
        self.expander_name = None
        self.creator = None
        self.function = None
        self.nodetype, self.template, self.function_name, self.code, self.lineno = \
                  template_data[self.label]

        if self.nodetype == 'STUB':
            self.do_stub()
        elif self.nodetype == 'CREATOR':
            self.do_creator()
        elif self.nodetype == 'TEMPLATE':
            self.do_template()
        elif self.nodetype == 'FUNCTION':
            self.do_function()
        else:
            raise TypeError, 'illegal code for template: %s' % self.nodetype

    def do_stub(self):
        """Processes STUB entries in the templates file"""
        self.code = MetaNode.stub_template
        self.do_creator()

    def do_creator(self):
        """Converts the CODE section of a CREATOR into a Python function"""
        # self.code is the text for a function "outer" that is wrapped around
        # a creator and returns it.
        try:
            exec self.code in globals(), locals()
        except:
            sys.stdout.flush()
            print >> sys.stderr, 'While creating the _MetaNode for "%s", an error ' \
                'was found\nin the CODE section starting at line %i of the templates file.' % \
                     (self.label, self.lineno)
            print >> sys.stderr, 'Note: you need to subtract one from the number in the traceback.'
            raise
        self.creator = outer()
        self.expander_name = 'EXcreator'

    def do_template(self):
        """Processes TEMPLATE entries in the templates file"""
        self.expander_name = 'EXtemplate'

    def do_function(self):
        """Converts the CODE part of a FUNCTION into a Python function"""
        try:
            exec self.code in globals(), locals()
        except:
            sys.stdout.flush()
            print >> sys.stderr, 'Error found in the following code ' \
                'from the templates file:'
            print >> sys.stderr, self.code
            raise
        Node.__dict__[self.label] = outer()

    def prepare_repr(self, attribute_name):
        attr = getattr(self, attribute_name)
        # Sometimes useful during debugging:
        # attr = getattr(self, attribute_name, 'attribute not present')
        if hasattr(attr, 'name'):
            attr = attr.name
        return '        %s: %s' % (attribute_name, repr(attr))

    def __repr__(self):
        out = ['<MetaNode:']
        out.append('    %s' % self.label)
        out.append('        nodetype: %s' % self.nodetype)
#        out.append(self.prepare_repr('treedata'))
        out.append(self.prepare_repr('expander'))
        out.append(self.prepare_repr('creator'))
        out.append('>')
        return '\n'.join(out)

class Node:
    """Each instance of <@...@> in a template becomes a node"""
    def __init__(self, label, tree, parent, start=None, end=None, indent=None,
                  data=None):
        # The "label" passed to Node must be:
        #     label ::= base_label ( . num_label )?
        #     base_label ::= letter (letter | digit | "_")*
        #     num_label ::= (digit)+
        # Examples: "a1_b", "a1_b.123"
        check_label(label)
        self.tree = tree
        self.all = tree.all
        self.parent = parent
        self.label = label
        self.base_label, self.num_label = parse_label(self.label)
        self.nodenum = tree.nodecounter
        tree.nodecounter += 1
        self.start = start
        self.end = end
        self.indent = indent

        # Make a copy here so "messages" dictionaries higher in the
        # node tree are not changed lower down.
        if self.parent is None:
            self.messages = dict()
        else:
            self.messages = dict(self.parent.messages)

        if not self.tree.metanodes.has_key(self.base_label):
            raise Exception, 'no entry for Node "%s" in the templates file' % self.label
        self.metanode = self.tree.metanodes[self.base_label]

        # Pass information from the MetaNode.
        self.lineno = self.metanode.lineno
        self.nodetype = self.metanode.nodetype
        self.template = self.metanode.template
        self.treedata = self.tree.treedata
        self.extDef = self.tree.treedata.extDef
        if self.metanode.expander_name == 'EXtemplate':
            self.expander = self.EXtemplate
        elif self.metanode.expander_name == 'EXcreator':
            self.expander = self.EXcreator

        self.creator = self.metanode.creator

        # The following will be filled in by "template_parser".
        self.subnode_list = []

        # "self.text" will contain the Pyrex text for the node.
        self.text = None

        # There are cyclic references within each Tree and its associated
        # MetaNodes and Nodes.
        self.tree.nodes[self.label] = self

        # global messages can be received only by nodes with two part names.
        if '.' in self.label:
            self.get_global_message()
        self.expander()

    # Utilities used in CODE in the templates file are placed here.
    def typespec(self, spec):
        newspec = []
        for s in spec:
            if s not in ('struct', 'union', 'enum'):
                newspec.append(s)
        return ' '.join(newspec)

    def text_assembler(self, depth=0):
        """Replaces the "<@...@>" parts of a template with text.

        Called once in "externalDef.py" and recursively here.
        """
        text = self.template
        if DEBUG:
            if depth == 0:
                print '----'
            print '* '*depth + 'enter text_assembler:', self.label, ':', text.replace('\n', '\\n')
            print '* '*depth + '             parents:', self.parents_labels()
        for subnode in self.subnode_list:
            if isinstance(subnode.text, str):
                subtext = subnode.text
            else:
                subtext = subnode.text_assembler(depth+1)

            # Now do the "reindent".
            if subnode.indent in (None, 0):
                replacement = subtext
            else:
                replacement = self.reindent(subtext, subnode.indent)

            text = '%s%s%s' % \
              (text[ : subnode.start], replacement, text[subnode.end : ])
        if DEBUG:
            print '* '*depth + 'exit text_assembler:', text.replace('\n', '\\n')
        return text

    def EXtemplate(self):
        """Expander for TEMPLATE nodes"""
        if DEBUG: print 'template:', self.parents_labels()
        self.template_parser()

    def EXcreator(self):
        """Expander for CREATOR nodes"""
        if DEBUG: print 'creator:', self.parents_labels()
        # Note the unusual call.
        self.template = self.creator(self)
        self.template_parser()

    def _prepare_repr(self, attribute_name):
        attr = getattr(self, attribute_name)
        # Use the next line instead of the previous line when debugging.
#        attr = getattr(self, attribute_name, 'attribute not present')
        if hasattr(attr, 'name'):
            attr = attr.name
        return '        %s: %s' % (attribute_name, repr(attr))

    def __repr__(self):
        out = ['<Node:']
        out.append('    %s' % self.label)
        out.append('        nodetype: %s' % self.nodetype)
        out.append(self._prepare_repr('text'))
        out.append(self._prepare_repr('template'))
#        out.append(self._prepare_repr('treedata'))
        out.append(self._prepare_repr('messages'))
        out.append('>')
        return '\n'.join(out)

    def parents(self):
        """Find the chain of parents for this node"""  
        plist = []
        nod = self
        while nod is not None:
            plist.append(nod)
            nod = nod.parent
        plist.reverse()
        return plist

    def parents_labels(self):
        """Find the labels for the chain of parents for this node"""
        plist = []
        nod = self
        while nod is not None:
            plist.append(nod.label)
            nod = nod.parent
        plist.reverse()
        return plist

    def template_parser(self):
        """Parses templates with "<@...@>" in them."""
        for start, end, label, indent, counter in self.template_labels():
            newnode = Node(label, self.tree, self, start, end, indent)
            self.subnode_list.append(newnode)
        # The subnode_data will be processed in reverse order so that text
        # replacements don't mess up "start" and "end".
        self.subnode_list.reverse()

    def template_labels(self):
        """Parses templates

        Find the "<@...@>" where "..." is some name.  If "<@...@>" is
        preceeded, on a line, only by spaces, all the lines in the
        replacement text will be indented the same as "<@...@>".
        """
        if not isinstance(self.template, str):
            raise TypeError, 'trying to find <@...@> in a template that ' \
                'is not a string: %s' % self.label
        counter = 0
        start = 0
        end = 0
        while 1:
            start = self.template.find('<@', end)
            if start == -1:
                return
            end = self.template.index('@>', start) + 2
            at_string = self.template[start:end]
            label = at_string[2:-2]

            # Is the "<@" preceded only by spaces on a line?
            i = start - 1
            while self.template[i] == ' ' and i >= 0:
                i -= 1
            if i == -1 or self.template[i] == '\n':
                indent = start - i - 1
            else:
                indent = None
            yield start, end, label, indent, counter
            counter += 1

    # The simple function "reindent" and two rules makes (almost) all
    # indentation problems disappear:
    #
    #   The pieces of text generated by this program never start or end with
    #   "\n"s. But they may contain "\n"s.
    #
    #   If "<@...@>" is preceeded, on a line of the template, only by
    #   spaces, all the lines in the replacement text will be indented the
    #   same as "<@...@>".
    #
    # If a line in a template consists of zero or more spaces followed by a
    # "<@...@>" and the "<@...@>" is replaced by "", then an unnecessary
    # line will be in the output.
    def reindent(self, text, indent):
        """Reindent a piece of text"""
        lines = text.splitlines()
        sep = '\n' + ' '*indent
        return sep.join(lines)

    # The next three functions are used to send information down the node
    # tree.
    def send_message(self, label, obj):
        """Send a message down the node tree"""
        self.messages[label] = obj

    def get_message(self):
        """Get a message.

        A uniqueified Node "xxx.1" can access information sent to "xxx" but
        not vice versa.
        """
        id, num = parse_label(self.label)
        if not self.label in self.messages and \
                 not self.base_label in self.messages:
            sys.stdout.flush()
            raise KeyError, 'No message found for %s in\n%s\nParents are %s' % \
                    (self.label, self.messages, self.parents_labels())
        if self.label in self.messages:
            return self.messages[self.label]
        else:
            return self.messages[self.base_label]

    def del_message(self):
        """Delete a message"""
        del self.messages[self.label]

    # A global message is used to send something to some specific node. The
    # receiver node must have a 2-part name (abcd.123). The object which is
    # sent is a dictionary of messages which is merged with the messages
    # dictionary of the receiving node. If "override" is True, these
    # messages override messages in the receiving dictionary.
    def send_global_message(self, label, message_dict, override=True):

        """A global message is used to send a message to some specific node.

        The receiver node must have a 2-part name (abcd.123).
        """
        check_label(label)
        if label.count('.') == 0:
            raise LabelError, \
             'nodes whose label is without a "." cannot process global messages: %s' % \
                 self.label
        if label in self.tree.global_messages:
            raise KeyError, 'global message to "%s" has already been sent from "%s"' % \
                   (label, self.label)
        if not isinstance(message_dict, dict):
            raise TypeError, 'global message must be a message dictionary'
        message_dict[None] = override
        self.tree.global_messages[label] = dict(message_dict)

    def get_global_message(self):
        """Get a global message"""
        if self.label.count('.') != 1:
            raise KeyError, \
             'nodes whose label is without a "." cannot process global messages: %s' % \
                 self.label
        if not self.label in self.tree.global_messages:
            sys.stdout.flush()
            raise KeyError, \
               'No global message found for %s in\n%s\nParents are %s' % \
                    (self.label, self, self.parents_labels())
        message_dict = self.tree.global_messages[self.label]
        if message_dict[None]:  # override
            self.messages.update(message_dict)
        else:
            for key in message_dict:
                if key not in self.messages:
                     self.messages[key] = message_dict[key]
        del self.tree.global_messages[self.label]

    def del_global_message(self):
        """Delete a global message"""
        if self.label.count('.') != 1:
            raise KeyError, \
             'nodes whose label is without a "." cannot process global messages: %s' % \
                 self.label
        del self.tree.global_messages[self.label]
