#! /usr/bin/env python

import os, sys
import Node
from Node import CODE, WALKAROUND_CODE
from myexceptions import *
from LN import LN, MLLN

class SPECTOKENS(object):
    OPEN = '<<OPEN>>'
    CLOSE = '<<CLOSE>>'
    NEWLINE = '<<NEWLINE>>'
    TAB = '<<TAB>>'
    SpecTokens = (NEWLINE, TAB, OPEN, CLOSE)

def CheckTokenListTypes(alist):
    """Lists the token codes in a tuple of tokens."""
    types = dict()

    def recurse(seq):
        for item in seq:
            t = type(item)
            if t in (tuple, list):
                recurse(item)
            else:
                if t not in types:
                    types[t] = None

    recurse(alist)
    if len(types) == 0:
            return 'empty'
    elif len(types) == 1:
        if Token in types:
            return 'Token'
        if str in types:
            return 'str'
    else:
        return types.keys()

def TokenTuple_AsString(tup):
    """Represents a tuple of tokens as an informative string."""

    def recurse(seq):
        out = []
        for item in seq:
            if isinstance(item, tuple):
                substring = '(%s)' % recurse(item)
                out.append(substring)
            elif isinstance(item, list):
                substring = '[%s]' % recurse(item)
                out.append(substring)
            else:
                if isinstance(item, Token):
                    item = item.asstring()
                out.append(repr(item))
        if isinstance(seq, tuple):
            return '(%s)' % ', '.join(out)
        elif isinstance(seq, list):
            return '[%s]' % ', '.join(out)
        else:
            return ', '.join(out)

    s = CheckTokenListTypes(tup)
    text = recurse(tup)
    return '[of "%s": %s]' % (s, text)

def Tuple_ToTokenTuple(text):
    """Converts an input string in ".data" file format into a tuple of tokens."""
    text = text.strip()
    words = text.splitlines()
    lastword = None
    out = []
    for word in words:
        if lastword == SPECTOKENS.OPEN:
            tok = Token(word, CODE.NODENAME)
        elif word == SPECTOKENS.OPEN:
            tok = Token(word, CODE.LP)
        elif word == SPECTOKENS.CLOSE:
            tok = Token(word, CODE.RP)
        else:
            tok = Token(word, CODE.CTOKEN)
        out.append(tok)
        lastword = word
    return tuple(out)

class Token(Node.Node):
    def __init__(self, string, code):
        """Creates a Token."""
        Node.Node.__init__(self)
        self.string = string
        self.code = code

    def copy(self):
        """Copies a token."""
        return Token(self.string, self.code)

    def reset(self):
        """Reinitializes a Token."""
        Node.Node.reset(self)

    def asstring(self):
        """Simplest sting describing a Token."""
        if self.code == 'OP':
            template = '%s'
        else:
            template = "'%s'"
        return template % self.string

    def is_open_parens(self):
        """Does this token contain a '('?"""
        return self.string == SPECTOKENS.OPEN and self.code == CODE.LP

    def is_close_parens(self):
        """Does this token contain a ')'?"""
        return self.string == SPECTOKENS.CLOSE and self.code == CODE.RP

    def get_tokens(self):
        out = []
        for token in self.walk():
            if token.code == CODE.CTOKEN:
                out.append(token.string)
        return out

    def walkaroundToken(self):
        """Walks around the outside of the tree structure headed by node.

        Each Node node with code == CODE.NODENAME is visited twice. The
        first visit is before its children are visited and corresponds to
        '('. At the start of the first visit, self.walkaround_code is set to
        CODE.LP. The second visit is after its children are visited and
        corresponds to ')'. Just before the second visit,
        self.walkaround_code is set to WALKAROUND_CODE.SECOND. Some
        NODENAMEs may be leaves. They are still visited twice.
        """
        walkaround_codes = dict()
        if self.code == CODE.NODENAME:
            walkaround_codes[self.id] = WALKAROUND_CODE.FIRST
            yield walkaround_codes[self.id], self
            for child in self.children:
                for code, node in child.walkaroundToken():
                    yield code, node
            walkaround_codes[self.id] = WALKAROUND_CODE.SECOND
            yield walkaround_codes[self.id], self
            del walkaround_codes[self.id]
        else:
            walkaround_codes[self.id] = WALKAROUND_CODE.LEAF
            yield walkaround_codes[self.id], self
            del walkaround_codes[self.id]
