#! /usr/bin/env python

import os, sys
import Token, Tree
from Node import CODE, WALKAROUND_CODE
from Token import SPECTOKENS
import strip_preprocs, unstrip_preprocs
from myexceptions import *
from LN import LN, MLLN

#>=== class TokenTree =============================================

class TokenTree(Tree.Tree):
    """Tree representation for a parser's output."""
    def __init__(self, data=None):
        """TokenTree is created using TokenTree(data), data optional."""
        Tree.Tree.__init__(self)
        # The spelling dict is used only in "pattern.py" on trees that have
        # been fully completed. It is updated only when trees that are 
        # likely to be complete are modified.
        self._speller = None
        if isinstance(data, str):
            data = Token.Tuple_ToTokenTuple(data)
        if isinstance(data, (tuple, list)):
            self.fromtuple(data)
            self.new_speller()
        elif isinstance(data, Token.Token):
            self.addnode(data, 'TOP')
        elif isinstance(data, Tree.Tree):
            Tree.Tree._copy(self, data)
        elif data is not None:
            raise TypeError, '"data" must be a string, list, tuple, Token, Tree, or None'

    def show_parens(self):
        """Like "__str__" but the '('s and ')'s are added."""
        out = []
        for walkaround_code, token in self.walkaround():
            inner = []
            inner.append(token.depth * '  ')
            if walkaround_code == WALKAROUND_CODE.FIRST:
                inner.append('(')
            inner.append(str(token))
            if walkaround_code == WALKAROUND_CODE.SECOND:
                inner.append(')')
            text = ''.join(inner)
            out.append(text)
        return '\n'.join(out)

    def new_speller(self):
        """Create the speller dict.

        Use below and in cases (such as "strip_special_tokens") where a
        completed tree is modified.
        """
        self._speller = dict()
        for node in self.top.walk():
            self._speller[node.string] = None

    def get_speller(self):
        if self._speller is None:
            LN('new speller')
            self.new_speller()
        else:
            LN('use old _speller')
        return self._speller
        speller = property(get_speller)

    def fromtuple(self, tokentuple):
        """Adds a tuple of Tokens to a new Tree."""
        if self.top is not None:
            raise ValueError, 'new or reset TokenTree required'
        if len(tokentuple) == 0:
            return
        i = 0
        # "child_bearer" is the node that children are currently being added
        # to.
        child_bearer = 'TOP'
        while i < len(tokentuple):
            if tokentuple[i].is_open_parens():
                i += 1  # Skip the '('.
                self.addnode(tokentuple[i], child_bearer)
                child_bearer = tokentuple[i]
                i += 1
            elif tokentuple[i].is_close_parens():
                if child_bearer == self.top:
                     return
                else:
                    child_bearer = child_bearer.parent
                i += 1
            else:  # C tokens.
                self.addnode(tokentuple[i], child_bearer)
                i += 1

    def combine(self):
        """Combines adjacent leaf nodes. Modifies the tree."""
        for token in self.reverse_walk():
            if token.isleaf():
                continue
            for i in range(len(token.children)-1, 0, -1):
                if token.children[i].code == CODE.CTOKEN and \
                       token.children[i-1].code == CODE.CTOKEN and \
                       token.children[i].string not in SPECTOKENS.SpecTokens and \
                       token.children[i-1].string not in SPECTOKENS.SpecTokens:
                    tok = token.children[i]
                    self.extract(token.children[i])
                    token.children[i-1].string = \
                         '%s %s' % (token.children[i-1].string, tok.string)

    def strip_special_tokens(self):
        """Removes '<<NEWLINE>>' and '<<TAB>>' Tokens from a TokenTree. Modifies the tree."""
        for token in self.reverse_walk():
            if not token.isleaf():
                continue
            if token.string in SPECTOKENS.SpecTokens:
                token.kill()
        self.new_speller()

    def get_tokens(self):
        return self.top.get_tokens()

    def select(self, selectdata, keep):
        """Removes selected tokens from a tree. Modifies the tree.

         "selectdata" can be a list, tuple, dict, or "RULEPRINTSONLY".
        """
        def will_keep(keep, nodename, selectdata):
            if selectdata == 'RULEPRINTSONLY':
                 base, ext = os.path.splitext(nodename)
                 return len(ext) == 0 or not ext[1:].isdigit()
            else:
                return (keep and nodename in selectdata) or \
                       (not keep and nodename not in selectdata)

        for token in self.reverse_walk():
            if not token.isleaf() and \
                    not will_keep(keep, token.string, selectdata):
                if token == self.top:
                    token.string = None
                else:
                    token.kill()
        self.new_speller()

    def to_string_list(self):
        """Produces a list of the stings in the Tokens of a TokenTree."""
        out = []
        first = True
        for token in self.walk():
            if not first:
                out.append(token.string)
            else:
                first = False
        return out

    def make_source(self):
        """Make compilable C code from a TokenTree. Messes up the TokenTree."""
        first, special, normal = 0, 1, 2
        self.stripNODENAMEleaves()
        self.select([], True)
        strings = self.to_string_list()
        out = []
        prev = first
        for i, string in enumerate(strings):
            if string == SPECTOKENS.NEWLINE:
                out.append('\n')
                prev = special
            elif string == SPECTOKENS.TAB:
                out.append('\t')
                prev = special
            else:
                if prev == normal:
                    out.append(' ')
                prev = normal
                out.append(string)
        out.append('\n')
        return ''.join(out)

    def reconstruct_text(self, LPisNone_sep, LPtoken_sep, LPindent,
              token_sep, indent, RPtoken_sep, RPindent, indent_size,
              evalable=False, suppress_special_tokens=False):
        """Converts a TokenTree into a useful and/or pretty string.

        LPisNone_sep   If the flattened form would start with "((",
                       put "LPisNone_sep" between the "("s.
        LPtoken_sep    Goes before all other "("s.
        LPindent       If True, put indentation after "LPtoken_sep".
        token_sep      Separates token strings.
        indent         If True, put indentation after "token_sep".
        RPtoken_sep    Goes before all  ")"s.
        RPindent       If True, put indentation after "RPtoken_sep".
        indent_size    Number of spaces per indent.
        evalable       Output will be evalable.
        suppress_special_tokens  If True, skip all the "<<NEWLINE>>" and
                                 "<<TAB>>" tokens

        An example is very useful. Create "tree" and run:

            tree.reconstruct_text(LPisNone_sep = '\nAA',
                  LPtoken_sep = ',\nBB', LPindent = True,
                  token_sep = ', \nCC', indent = True,
                  RPtoken_sep = 'DD', RPindent = False,
                  indent_size = 2, evalable = False,
                  suppress_special_tokens = False)

        The output is:

            (externalList,
            BB  (externalDef,
            BB    (declaration,
            BB      (declSpecifiers,
            BB        (typeSpecifier,
            BB          (typeSpecifier.3,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |<<NEWLINE>>|,
            CC            |int|DD)DD)DD),
            BB      (initDeclList,
            BB        (initDecl,
            BB          (declarator,
            BB            (declarator.0,
            CC              |k|DD)DD)DD)DD),
            CC      |;|DD)DD),
            ...
        """
        if self.top is None:
            return ''
        topsNone = self.top.string is None
        counter = 0
        out = []
        for walkaround_code, token in self.walkaround():
            if walkaround_code == WALKAROUND_CODE.FIRST:
                # We want nothing to be before the initial "(", so the
                # counter == 0 case is ignored.
                if counter == 1 and topsNone:
                    out.append(LPisNone_sep)
                elif counter > 0:
                    out.append(LPtoken_sep)
                if LPindent:
                    spaces = token.depth * indent_size * ' '
                    if not topsNone or counter != 1 or \
                               LPisNone_sep[-1] == '\n':
                        out.append(spaces)
                out.append('(')
                if token.string is not None:
                    if evalable:
                        out.append(repr(token.string))
                    else:
                        out.append(token.string)
            elif walkaround_code == WALKAROUND_CODE.SECOND:
                out.append(RPtoken_sep)
                if RPindent:
                    out.append(token.depth * indent_size * ' ')
                out.append(')')
            elif not suppress_special_tokens or \
                      token.string not in (SPECTOKENS.NEWLINE, SPECTOKENS.TAB):
                out.append(token_sep)
                if indent:
                    out.append(token.depth * indent_size * ' ')
                if evalable:
                    out.append(repr(token.string))
                else:
                    out.append("|%s|" % token.string)
            counter += 1
        out.append('\n')
        return ''.join(out)

    def as_string_pretty(self, flatRP=True, flatCtoken=True,
              evalable=False, suppress_special_tokens=False):
        """TokenTree pretty printer."""
        _LPisNone_sep = '\n'
        _LPtoken_sep = ',\n'
        _LPindent = True

        if flatRP:
            _RPtoken_sep = ''
            _RPindent = False
        else:
            _RPtoken_sep = '\n'
            _RPindent = True

        if flatCtoken:
            _token_sep = ', '
            _indent = False
        else:
            _token_sep = ',\n'
            _indent = True
        _indent_size = 2

        return self.reconstruct_text(_LPisNone_sep,
                  _LPtoken_sep, _LPindent,
                  _token_sep, _indent,
                  _RPtoken_sep,  _RPindent,
                  _indent_size, evalable,
                  suppress_special_tokens)

    def as_string_pretty2(self):
        return self.as_string_pretty(True, False, False, True)

    def asstring_flat(self, suppress_special_tokens=False):
        """Outputs a string which is the repr of Python nested tuple.

        Because the Python parser stack is too small, eval will usually
        fail on this string.
        """
        return self.reconstruct_text(LPisNone_sep = '',
                  LPtoken_sep = ', ', LPindent = False,
                  token_sep = ', ', indent = False,
                  RPtoken_sep = '', RPindent = False,
                  indent_size = 0, evalable=True,
                  suppress_special_tokens = suppress_special_tokens)

    def asstring_test(self, suppress_special_tokens=False):
        """Demo for "reconstruct_text"."""
        return self.reconstruct_text(LPisNone_sep = '\nAA',
                  LPtoken_sep = ',\nBB', LPindent = True,
                  token_sep = ', \nCC', indent = True,
                  RPtoken_sep = 'DD', RPindent = False,
                  indent_size = 2, evalable=False,
                  suppress_special_tokens = suppress_special_tokens)

    def to_datafile_format(self):
        """Outputs a string which is in the format for a ".data" file."""
        out = []
        for walkaround_code, token in self.walkaround():
            if walkaround_code == WALKAROUND_CODE.FIRST:
                out.append(SPECTOKENS.OPEN)
                out.append(token.string)
            elif walkaround_code == WALKAROUND_CODE.SECOND:
                out.append(SPECTOKENS.CLOSE)
            elif walkaround_code == WALKAROUND_CODE.LEAF:
                out.append(token.string)
        out.append('')
        return '\n'.join(out)

#>=== Functions that work with TokenTrees ==================================

def Combine(tree):
    """Combines adjacent leaf nodes. Returns a new tree."""
    return TokenTree(tree).combine()

def StripSpecialTokens(tree):
    """Removes '<<NEWLINE>>' and '<<TAB>>' Tokens from a TokenTree. Returns a new tree."""
    return TokenTree(tree).strip_special_tokens()

def StripNODENAME_Leaves(tree):
    """Remove leaves that are also NODENAMEs. Returns a new tree.

    Use this function only after the tree is completed.
    """
    # This new tree is not suitable for spelling checks.
    newtree = TokenTree(tree).stripNODENAMEleaves()

def Select(tree, selectdata, keep):
    """Removes selected tokens from a tree. A new tree is returned.

    "selectdata" can be a list, tuple or dict.
    """
    return TokenTree(tree).select(selectdata, keep)

def MakeSource(tree):
    """Make compilable C code from a TokenTree. Returns a new tree."""
    return TokenTree(tree).make_source()

def MakeNest(tree):
    return tree.as_string_pretty2()

def MakeNest_Files(datafile, nestfile):
    """Outputs the data from a ".data" file in a easily readable form."""
    text = open(datafile, 'r').read()
    tree = TokenTree(text)
    text = tree.as_string_pretty()
    open(nestfile, 'w').write(text)

def MakeSrc_Files(datafile, srcfile):
    """Outputs the data from a ".data" file as compilable C code."""
    text = open(datafile, 'r').read()
    tree = TokenTree(text)
    text = tree.make_source()
    file(srcfile, 'w').write(text)
