from __future__ import generators
import sys
from linenum import LN
from externalDef import DEBUG

"""
The first line of a template must start with one of the words TEMPLATE,
CREATOR, FUNCTION, or STUB. The rest of the line is a label. Any lines after
the first line are the template. An "rstrip()" is run on each non-empty
template. For CREATORs, there may also be a CODE section. The first line of
the CODE section is the word "CODE" followed by the name of a function. The
code for the function is on the remaining lines. So the general form is

<nodetype> <label>        (TEMPLATE, CREATOR, FUNCTION, or STUB)
<template>
CODE <function_name>
def <function_name>(node):
    <Python code>

The function in the CODE part of a CREATOR is called inside an instance of
Node to create a final template for the Node. If the <template> part of the
CREATOR is given, it can be accessed by the function as "node.template".

A typical CREATOR is:

CREATOR loop_1d
for LocaLi from 0 <= LocaLi < <@dim0@>:
    <@%s@>
CODE Creator_loop_1d
def Creator_loop_1d(node):
    label = node.get_message()
    return node.template % label

A TEMPLATE can be viewed as a CREATOR whose CODE is

   CODE funct
   def funct(node):
       return node.template

I.e., the template that is given is the final template.

"Stub name" becomes "<STUB: name>". Used for stubs.

A FUNCTION defines a function that can be called by the CODE for CREATORs.
"FUNCTION something" is accessed in a CREATOR by "node.something().
"""
def _templates(text):
    text = text.strip()
    # Lines starting with "@" are removed.
    lines = text.splitlines()
    newlines = []
    for line in lines:
        if len(line) == 0 or line[0] != '@':
            newlines.append(line)
    text = '\n'.join(newlines)
    splitpoints = []
    for word in ('TEMPLATE', 'CREATOR', 'STUB', 'FUNCTION'):
        start = 0
        while 1:
            start = text.find(word, start)
            if start == -1:
                break
            if start == 0 or text[start-1] == '\n':
                splitpoints.append(start)
            start += len(word)
    splitpoints.sort()

    splitpoints.append(len(text))
    for i in range(len(splitpoints) - 1):
        yield text[splitpoints[i] : splitpoints[i+1]].rstrip()

def _splitCODE(text):
    if text.find('\nCODE') == -1:
        return (text, None)
    return text.split('\nCODE')

def _parse_upper(upper):
    lines = upper.splitlines()
    if lines[0][-1] == ':':
        lines[0] = lines[0][:-1]
    words = lines[0].split()
    if len(lines) > 1:    
        rest = '\n'.join(lines[1:])
    else:
        rest = None
    return words[0], words[1], rest

def _parse_lower(lower):
    lines = lower.strip().splitlines()
    if len(lines) <= 1:
        raise Exception, '"CODE" present without any code'
    first = lines[0]
    if first[-1] == ':':
        first = first[:-1]
    function_name = first.strip()
    out = []
    out.append('def outer():')
    for line in lines[1:]:
        out.append('    %s' % line)
    out.append('    return %s' % first)
    out.append('')
    code = '\n'.join(out)
    return function_name, lines[1], code

def template_dict(text):
    template_data_dict = dict()
    labels = []
    for t in _templates(text):
        upper, lower = _splitCODE(t)
        entrytype, label, template = _parse_upper(upper)
        if lower is not None:
            function_name, def_line, code = _parse_lower(lower)
            k = text.index(def_line)
            lineno = text.count('\n', 0, k) + 1
        else:
            function_name = code = lineno = None
        template_data_dict[label] = (entrytype, template, function_name,
                        code, lineno)
        labels.append(label)
    return template_data_dict, labels
