#! /usr/bin/env python

from __future__ import generators

import sys
from linenum import LN

""" Matches a pattern against a tree.

Trees are represented by nested tuples. The first entry in the tuple is the
parent; the rest of the tuple are the children. The tree

   A
  / \   is represented by ("A", ("B", "b"), ("C", "c", "d"))
 B   C
 |  / \    
 b c   d

The pattern is a tuple with entry for each level of the data tree. So
("A", "C") has a match in the example above. Each entry in the pattern must
be a parent. The output of a match is the subtree headed by the last parent
in the pattern. For the example the output would be ("C", "c", "d").
the entries in a pattern can be
  A string naming a parent node.
  ANY which matches anything.
  (OR, X, Y, Z) which matches X, Y, or Z.
"""
class NoMatchException(Exception):
    pass

class _OR:
    def __repr__(self):
        return 'OR'
OR = _OR()

class _ANY:
    def __repr__(self):
        return 'ANY'

    def __eq__(self, obj):
        return True

    def __ne__(self, obj):
        return False

    def __lt__(self, obj):
        return NotImplemented

    def __le__(self, obj):
        return NotImplemented

    def __gt__(self, obj):
        return NotImplemented

    def __ge__(self, obj):
        return NotImplemented
ANY = _ANY()

#>=== walktuple =================================================

def walktuple(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple. """
    yield top
    for obj in top:
        if isinstance(obj, tuple):
            for obj2 in walktuple(obj):
                yield obj2

def walktuple_all(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple or token. """
    yield top
    for obj in top:
        if isinstance(obj, tuple):
            for obj2 in walktuple_all(obj):
                yield obj2
        else:
            yield obj

def walktuple_tokens(top):
    """ Depth-first walk through a nested tuple. Yields each token. """
    for obj in walktuple_all(top):
        if not isinstance(obj, tuple):
            yield obj

def walktuple_depth(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple and its depth."""
    def inner(top, depth):
        yield top, depth
        for obj in top:
            if isinstance(obj, tuple):
                for obj2, depth2 in inner(obj, depth+1):
                    yield obj2, depth2

    for obj, depth in inner(top, 0):
        yield obj, depth

def walktuple_parents(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple and its list of parents."""
    def inner(top, parents):
        yield top, parents
        for obj in top:
            if isinstance(obj, tuple):
                p = list(parents)  # Force copy.
                p.append(obj[0]) 
                for obj2, parents2 in inner(obj, p):
                    yield obj2, parents2

    for obj, parents in inner(top, [top[0]]):
        yield obj, parents

def walktuple_depth_all(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple or token and its depth."""
    def inner(top, depth):
        yield top, depth
        for obj in top:
            if isinstance(obj, tuple):
                for obj2, depth2 in inner(obj, depth+1):
                    yield obj2, depth2
            else:
                yield obj, depth

    for obj, depth in inner(top, 0):
        yield obj, depth

#>=== prettyprint =======================================================

class _FIRST:
    def __repr__(self):
        return '<FIRST>'
FIRST = _FIRST()

class _LAST:
    def __repr__(self):
        return '<LAST>'
LAST = _LAST()

class _LASTone:
    def __repr__(self):
        return '<LASTone>'
LASTone = _LASTone()

def walktuple_xtras(top):
    """ Depth-first walk through a nested tuple. Yields each subtuple or token.

    Also yields the "("s and ")"s as "FIRST" and "LAST".
    """
    def inner(top, depth):
        yield FIRST, depth
        for obj in top:
            if isinstance(obj, tuple):
                for obj2, depth2 in inner(obj, depth+1):
                    yield obj2, depth2
            else:
                yield obj, depth
        if len(top) == 1:
            yield LASTone, depth
        else:
            yield LAST, depth

    for obj, depth in inner(top, 0):
        yield obj, depth

def prettyprint(top, indent=3):
    """ Prettyprints a nested tuple"""
    sFIRST = str(FIRST)
    sLAST = str(LAST)
    sLASTone = str(LASTone)
    out = []
    lastdepth = -1
    for obj, depth in walktuple_xtras(top):
        spaces = ' ' * indent * depth
        if obj == FIRST:
            s = '\n' + spaces + sFIRST
        elif obj == LAST:
            s = sLAST + ', '
            # Remove any comma just before the last ")" of a tuple unless
            # the tuple has only one member.
            if out[-1].endswith(', '):
                out[-1] = out[-1][:-2]
        elif obj == LASTone:
           s = sLASTone + ', '
        else:
            if out[-1].endswith(sLAST + ', '):
                com = '\n' + spaces
            else:
                com = ''
            s = '%s"%s", ' % (com, str(obj))
        out.append(s)
    s =''.join(out)
    s = s.replace(sLAST + '\n', sLAST + ',\n')
    s = s.replace(sLASTone + '\n', sLASTone + ',\n')
    s = s.replace(sFIRST, '(')
    s = s.replace(sLAST, ')')
    s = s.replace(sLASTone, ')')
    return s[1:-2]

def prettyprint_source(tokenlist):
    newlist = []
    for token in tokenlist:
        if token in ('nl1', 'nl2', 'nl3', 'nl4'):
            newlist.append('\n')
        elif token == 'SEMI':
            newlist.append(';')
        elif token == 'COMMA':
            newlist.append(',')
        elif token == 'LPARENS':
            newlist.append('(')
        elif token == 'RPARENS':
            newlist.append(')')
        else:
            newlist.append(token)
    newlist2 = []
    for token in newlist:
        if token in ('\n', '\t'):
            newlist2.append(token)
        else:
            newlist2.append(token + ' ')
    return ''.join(newlist2)

def walkhist(top, maxdepth=None):
    """ Depth-first walk through a nested tuple where a list of nodes above the current one is kept.

    If the tree is

        ('declaration',
            ('storageClassSpecifier', 'typedef'),
            ('unionSpecifier', 'union',
                ('structOrUnionBody', '{', ...)
            ...


    the hists are

        ['declaration']                                              depth = 0
        ['declaration', 'storageClassSpecifier']                     depth = 1
        ['declaration', 'storageClassSpecifier', 'typedef']          depth = 1
        ['declaration', 'unionSpecifier']                            depth = 1
        ['declaration', 'unionSpecifier', 'union']                   depth = 1
        ['declaration', 'unionSpecifier', 'structOrUnionBody']       depth = 2
        ['declaration', 'unionSpecifier', 'structOrUnionBody', '{']  depth = 2
        ...

    Each hist sometimes ends in a token from the input source code:
    "struct", "CvInput", "{", etc. If this token is removed, then
    depth = len(hist) - 1.

    Currently hists never ends in a token from the input source code. See
    below.

    Note for the future. Types of antlr matching rules.
       A B C           sequence
       A | B | C       or match
       (A)             subrule
       (A)*            zero or more
       (A)+            one or more
       (A)?            zero or one
    """
    def inner(top, maxdepth, hist):
        if len(hist) > maxdepth+1:
            return
        first = 0
        last = len(top)
        if top[0] == 'TLMB827':
            first = 1
            last = len(top) - 1
        newhist = hist + [top[first]]  # Forces a copy to be made.
        yield top, newhist
        for obj in top[first+1:last]:
            if isinstance(obj, tuple):
                if len(hist) < maxdepth:
                    for obj2, hist2 in inner(obj, maxdepth, newhist):
                        yield obj2, hist2
#            else:
#                # Tokens from the source code may be at the end of a hist.
#                yield obj, newhist + [obj]

    if maxdepth is None:
         maxdepth = 2**31
    for obj, hist in inner(top, maxdepth, []):
        yield obj, hist

#>=== match ========================================================

def match_last(top, patt):
    if not isinstance(top, tuple):
        return False
    if not isinstance(patt[-1], tuple):
        return ( top[0] == patt[-1] )
    return ( top[0] in patt[-1][1:] )

def match(top, patt):

    def inner(top, depth):
        if depth == len(patt) - 1:
            if match_last(top, patt):
                yield top
            else:
                return
        elif isinstance(patt[depth+1], tuple) and patt[depth+1][0] == OR:
            for obj in top[1:]:
                for s in patt[depth+1][1:]:
                    if isinstance(obj, tuple) and obj[0] == s:
                        for obj2 in inner(obj, depth+1):
                            yield obj2
        else:
            for obj in top[1:]:
                if isinstance(obj, tuple) and obj[0] == patt[depth+1]:
                    for obj2 in inner(obj, depth+1):
                        yield obj2

    if not isinstance(top, (tuple, list)):
        raise TypeError, 'first parameter not a list or tuple: %s' % top
    for obj in inner(top, 0):
        yield obj

def all_matches(top, patt):
    if type(top) not in (tuple, list):
        raise TypeError, 'first parameter not a list or tuple: %s' % top
    out = []
    for obj in match(top, patt):
        out.append(obj)
    return tuple(out)

def first_match(top, patt):
    """ Return first match found between "top" and "patt". """
    for obj in match(top, patt):
        return obj
    return None

def first(top):
    """ Returns top[1]"""
    if type(top) not in (tuple, list):
        raise TypeError, 'parameter not a list or tuple: %s' % top
    if len(top) < 2:
        raise NoMatchException, 'not enough elements'
    return top[1]

def nth(top, n):
    """ Returns top[n]"""
    if type(top) not in (tuple, list):
        raise TypeError, 'first parameter not a list or tuple: %s' % top
    if len(top) < n+1:
        raise NoMatchException, 'not enough elements'
    return top[n]
        
def gettokens(top):
    """ Return a list of all tokens found in a nested tuple """
    if isinstance(top, str):
        return (top,)
    out = []
    for obj in top[1:]:
        if isinstance(obj, str):
            out.append(obj)
        else:
            out2 = gettokens(obj)
            out.extend(out2)
    return tuple(out)

def gettokens_join(top):
    tokens = gettokens(top)
    tokens = replace_spectokens(tokens)
    return ' '.join(tokens).strip()

# BUG? What about 'nl1', etc.?
def replace_spectokens(tokens):
    d = { 'COMMA': ',',
          'LPARENS': '(',
          'RPARENS': ')' }
    out = []
    for i in range(len(tokens)):
        if i > 0 and i < len(tokens)-1 and \
             tokens[i-1:i+2] == ['LPARENS', 'void', 'RPARENS']:
            continue
        if tokens[i] == 'const':
            continue
        if d.has_key(tokens[i]):
            out.append(d[tokens[i]])
        else:
            out.append(tokens[i])
    return out

if __name__ == '__main__':
#    data = ('A', ('H', 'd', ('J', 'j')), ('B', 'b'), ('C', 'x'))
#    pattern = ('A', (OR, 'B', 'C'))
    data = ('A', ('H', 'd', ('J', 'j')), ('B', 'b'), ('C', ('X', 'u')))
    pattern = ('A', (OR, 'B', 'C'), 'X')
    print data
    for obj in match(data, pattern):
        print '==>', obj
    for obj, parents in walktuple_parents(data):
        print obj, parents
    """
    data = eval(file('nano.clean', 'r').read())
    pattern1 = ("externalDef", "declaration", "declSpecifiers", "typeSpecifier")
    obj1 = first_match(data[1][0], pattern1)
    pattern = ("typeSpecifier", (OR, "structSpecifier", "unionSpecifier"))
    for obj2 in match(obj1, pattern):
        print '==>', obj2
    """
    """
    data = (('A',),)
    print prettyprint(data)
    """
